#!/bin/sh

# Copyright (c) 2015, Electric Cloud, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   - Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#
#   - Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# (As an historical note, the above license instantiates
# the "BSD 2-Clause License" template that was presented
# by the Open Source Initiative at
#
#   http://www.opensource.org/licenses/bsd-license.php
#
# on 2012-02-23.)

# Summary:
#
# Takes one argument, a pathname.  Writes to that pathname
# a header file that defines macros describing various
# peculiarities of the current programming environment.
#
# Any test compilations are performed by:
#
#   $CC $CFLAGS $CPPFLAGS
#

set -e

if test $# -eq 0 -o _"$1" = _"-h"
then
    echo "Usage: $0 <output-header> <log> <compiler> [<compiler-arg> ...]" >&2
    exit 2
fi

out="$1"
shift

log="$1"
shift

tmpout="$out.$$"
tmpc="/tmp/genconfig$$.c"
trap 'rm -rf "$tmpout" "$tmpc"' EXIT HUP INT TERM QUIT

rm -f "$log" "$out" "$tmpout" "$tmpc"


testCompile() {
    echo "==== probe:" >> "$log"
    echo "$probe" >> "$log"
    echo "" >> "$log"

    echo "---- source:" >> "$log"
    cat "$tmpc" >> "$log"
    echo "" >> "$log"

    echo "---- compilation:" >> "$log"
    if "$@" -c -Wall -Werror -o /dev/null "$tmpc" >> "$log" 2>&1
    then
        success=1
    else
        success=0
    fi
    echo "" >> "$log"
}

logResult() {
    echo "---- result:" >> "$log"
    echo "$result" >> "$log"
    echo "" >> "$log"
    echo "" >> "$log"
}


# POSIX_STRERROR_R is 1 if strerror_r has the function
# signature that is specified by POSIX; otherwise it is 0.

probe=POSIX_STRERROR_R

cat > "$tmpc" <<'EOF'
#define _POSIX_C_SOURCE 200809L
#define _XOPEN_SOURCE   500
#include <string.h>
int (*p)(int, char *, size_t) = &strerror_r;
EOF

testCompile "$@"

result="#define $probe $success"
echo "$result" >> "$tmpout"

logResult


# FDOPENDIR_SUPPORTED is 1 if fdopendir is supported; otherwise it is 0.

probe=FDOPENDIR_SUPPORTED

cat > "$tmpc" <<'EOF'
#define _POSIX_C_SOURCE 200809L
#define _XOPEN_SOURCE   500
#include <dirent.h>
DIR *f(int fd) { return fdopendir(fd); }
EOF

testCompile "$@"

result="#define $probe $success"
echo "$result" >> "$tmpout"

logResult


mv "$tmpout" "$out"
