/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * These are white-box tests of the implementation of vrpsx_embedding_t
 * defined by the parent directory (but not of its vrpsx_system_t portion).
 *
 * By "white-box" we mean that these tests are designed for one specific
 * implementation, and may need adjustment when that implementation changes,
 * even if those changes are backward-compatible at the interface level.
 *
 * In principle one could also define "black-box" tests that require only
 * those behaviors guaranteed by the interface itself, though perhaps with
 * the restriction that they be applied to a particular execution environment.
 * As yet we have not defined such tests, and therefore these white-box tests
 * may overlap what such black-box tests would do.  Feel free to separate out
 * white-box versus black-box testing, if you are of a mind to do so.
 */

#define _POSIX_C_SOURCE 200809L

#include "test_allocator.h"
#include "test_error.h"
#include "test_timer.h"
#include <vrpsx_embedding_main.h>
#include <vrpsx_embedding_main_identifier.h>
#include <vrpsx_identified_function.h>
#include <vrpsx_identified_function_identifier.h>
#include <vrpsx_utility.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <time.h>


#define MALLOC(TYPE)  ((TYPE *)malloc(sizeof(TYPE)))

#define ZERO(PTR)     (memset((PTR), 0, sizeof *(PTR)))


#if defined(__CYGWIN__)

static char *cygroot = NULL;

static void cygroot_initialize(void)
{
    int pfd[2];
    int ern = pipe(pfd);
    if (ern) {
        perror("pipe");
        exit(1);
    }

    pid_t child = fork();
    if (child == -1) {
        perror("fork");
        close(pfd[0]);
        close(pfd[1]);
        exit(1);
    } else if (child == 0) {
        close(pfd[0]);
        dup2(pfd[1], 1);
        close(pfd[1]);

        int fd = open("/dev/null", O_RDONLY);
        dup2(fd, 0);
        close(fd);

        static char *argv[] = { "cygpath", "-w", "/", NULL };
        execvp(argv[0], argv);
        perror(argv[0]);
        exit(1);
    } else {
        close(pfd[1]);

        cygroot = (char *)malloc(4096);
        size_t size = 0;

        while (size < 4096 - 1) {
            ssize_t num_read = read(pfd[0], cygroot + size, 4096 - 1 - size);
            if (num_read == (ssize_t)-1) {
                perror("read");
                exit(1);
            }
            if (num_read == 0) {
                break;
            }
            size += num_read;
        }
        close(pfd[0]);
        if (size && cygroot[size - 1] == '\n') {
            --size;
        }
        cygroot[size] = '\0';

        int status = -1;
        if (waitpid(child, &status, 0) != child) {
            perror("waitpid");
            exit(1);
        }

        if (! WIFEXITED(status) || WEXITSTATUS(status)) {
            fprintf(stderr, "cygpath status: %d\n", status);
            exit(1);
        }
    }
}

static char *cygroot_relative(const char *tail)
{
    size_t root_length = strlen(cygroot);
    size_t tail_length = strlen(tail);
    size_t separate = (root_length != 0 && cygroot[root_length - 1] != '\\')
            && (tail_length != 0 && tail[0] != '\\');
    char *result = malloc(root_length + separate + tail_length + 1);
    memcpy(result, cygroot, root_length);
    if (separate) {
        result[root_length] = '\\';
    }
    memcpy(result + root_length + separate, tail, tail_length);
    result[root_length + separate + tail_length] = '\0';
    return result;
}

static char *execution_path_concat(const char *path1, const char *path2)
{
    size_t length1 = strlen(path1);
    size_t length2 = strlen(path2);
    char *result = malloc(length1 + 1 + length2 + 1);
    memcpy(result, path1, length1);
    result[length1] = ';';
    memcpy(result + length1 + 1, path2, length2);
    result[length1 + 1 + length2] = '\0';
    return result;
}

#endif /* defined(__CYGWIN__) */


static void printEError(
    where_t where,
    vrpsx_embedding_t *embedding,
    char const *operation,
    vrpsx_errno_t ern)
{
    printError(
        where,
        &embedding->system,
        operation,
        ern);
}


#if defined(__CYGWIN__)

static char *skipHorizontalSpace(char *cp)
{
    while (isspace((unsigned char)*cp) && *cp != '\n') {
        ++cp;
    }
    return cp;
}

static char *endOfWord(char *cp)
{
    while (! isspace((unsigned char)*cp) && *cp) {
        ++cp;
    }
    return cp;
}

static int parseLong(char *begin, char *end, long *outValue)
{
    char oldEnd;
    char *foundEnd;

    if (begin == end) {
        *outValue = 0;
        return -1;
    }

    oldEnd = *end;
    *end = '\0';

    errno = 0;
    *outValue = strtol(begin, &foundEnd, 0);

    *end = oldEnd;

    if (errno || foundEnd != end) {
        *outValue = 0;
        return -1;
    }

    return 0;
}

static char *run(char **argv)
{
    int doomed, null = -1, fd[2] = { -1, -1 };
    pid_t child = -1;
    int status = 0;
    char *newBuffer, *buffer = NULL;
    size_t size = 0;
    size_t capacity = 0;
    ssize_t nread;

    fflush(stdout);
    fflush(stderr);

    null = open("/dev/null", O_RDONLY);
    if (null == -1) {
        perror("open");
        goto fail;
    }

    if (pipe(fd)) {
        fd[0] = -1;
        fd[1] = -1;
        perror("pipe");
        goto fail;
    }

    child = fork();
    if (child == -1) {
        perror("fork");
        goto fail;
    }
    if (child == 0) {
        if (close(fd[0])) {
            perror("close");
            goto child_fail;
        }
        if (dup2(null, 0) == -1) {
            perror("dup2");
            goto child_fail;
        }
        if (close(null)) {
            perror("close");
            goto child_fail;
        }
        if (dup2(fd[1], 1) == -1) {
            perror("dup2");
            goto child_fail;
        }
        if (close(fd[1])) {
            perror("close");
            goto child_fail;
        }
        execvp(argv[0], argv);
        perror("execvp");
    child_fail:
        _exit(EXIT_FAILURE);
    }

    doomed = fd[1];
    fd[1] = -1;
    if (close(doomed)) {
        perror("close");
        goto fail;
    }

    doomed = null;
    null = -1;
    if (close(doomed)) {
        perror("close");
        goto fail;
    }

    for (;;) {
        if (size >= capacity) {
            capacity = (capacity <= 512) ? 1024 : (capacity << 1);
            newBuffer = realloc(buffer, capacity);
            if (! newBuffer) {
                goto fail;
            }
            buffer = newBuffer;
        }

        nread = read(fd[0], buffer + size, capacity - size);
        if (nread == -1) {
            perror("read");
            goto fail;
        }
        if (nread == 0) {
            break;
        }
        size += nread;
    }

    doomed = fd[0];
    fd[0] = -1;
    if (close(doomed)) {
        perror("close");
        goto fail;
    }

    if (waitpid(child, &status, 0) == -1) {
        child = -1;
        perror("waitpid");
        goto fail;
    }
    child = -1;

    if (! WIFEXITED(status)) {
        fprintf(stderr, "child process exited abnormally\n");
        goto fail;
    }
    if (WEXITSTATUS(status) != 0) {
        fprintf(stderr, "child process exited with code %d\n",
                WEXITSTATUS(status));
        goto fail;
    }

    newBuffer = realloc(buffer, size + 1);
    if (! newBuffer) {
        goto fail;
    }
    buffer = newBuffer;
    buffer[size] = '\0';

    return buffer;

fail:
    if (buffer) {
        free(buffer);
    }
    if (null != -1) {
        close(null);
    }
    if (fd[0] != -1) {
        close(fd[0]);
    }
    if (fd[1] != -1) {
        close(fd[1]);
    }
    if (child != -1) {
        kill(child, SIGTERM);
        waitpid(child, &status, 0);
    }
    return NULL;
}

#endif /* defined(__CYGWIN__) */


int test_internalize_pid(
    vrpsx_embedding_t *embedding)
{
#if defined(__CYGWIN__)

    static char *argv[] = {
        "/usr/bin/ps",
        NULL
    };
    char *ps = 0, *ln, *pidBegin, *pidEnd, *winPidBegin, *winPidEnd, *cp;
    long pid, winPidL;
    vrpsx_pid_t actual;
    vrpsx_uint32_t winPid;
    vrpsx_errno_t ern;

        ps = run(argv);
    if (! ps) {
        goto fail;
    }

    for (ln = ps; (ln = strchr(ln, '\n')) != NULL;) {
        ++ln;

        pidBegin = skipHorizontalSpace(ln);
        if (! *pidBegin) {
            break;
        }

        /* Some lines of Cygwin "ps" output are prefixed with "I".
         * Skip that word if present.
         */
        if (! isdigit((unsigned char)*pidBegin)) {
            pidBegin = endOfWord(pidBegin);
            pidBegin = skipHorizontalSpace(pidBegin);
        }

        pidEnd = endOfWord(pidBegin);
        if (parseLong(pidBegin, pidEnd, &pid)) {
            PRINT_HERE;
            printf("bad Cygwin PID: '%.*s'\n",
                    (int)(pidEnd-pidBegin), pidBegin);
            goto fail;
        }

        cp = skipHorizontalSpace(pidEnd);
        cp = endOfWord(cp);

        cp = skipHorizontalSpace(cp);
        cp = endOfWord(cp);

        winPidBegin = skipHorizontalSpace(cp);
        winPidEnd = endOfWord(winPidBegin);
        if (parseLong(winPidBegin, winPidEnd, &winPidL)) {
            PRINT_HERE;
            printf("bad Windows PID: '%.*s'\n",
                    (int)(winPidEnd-winPidBegin), winPidBegin);
            goto fail;
        }
        winPid = (vrpsx_uint32_t)winPidL;

        ern = VRPSX_CALL3(embedding, internalize_pid,
                &winPid,
                sizeof winPid,
                &actual);
        if (ern == VRPSX_CONSTANT(&embedding->system, esrch) && kill(pid, 0)) {
            /* Process no longer exists; that is OK. */
            continue;
        }
        if (ern) {
            printEError(HERE, embedding, "internalize_pid", ern);
            goto fail;
        }
        if (actual != pid) {
            PRINT_HERE;
            printf("pid %ld != pid %ld\n", (long)actual, (long)pid);
            goto fail;
        }
    }

    free(ps);
    return 0;

fail:
    if (ps) {
        free(ps);
    }
    return -1;

#else /* not Cygwin */

    pid_t expected = getpid();
    vrpsx_pid_t actual;
    vrpsx_errno_t ern;

    ern = VRPSX_CALL3(embedding, internalize_pid,
            &expected,
            sizeof expected,
            &actual);
    if (ern) {
        printEError(HERE, embedding, "internalize_pid", ern);
        return -1;
    }
    if (actual != expected) {
        PRINT_HERE;
        printf("pid %ld != pid %ld\n", (long)actual, (long)expected);
        return -1;
    }

    return 0;

#endif /* Cygwin? */
}


int test_externalize_pid(
    vrpsx_embedding_t *embedding)
{
    test_allocator_t allocator;
    vrpsx_pid_t expected = getpid();
    void *outer_pid = (void *)0xBAD;
    size_t outer_size;
    vrpsx_errno_t ern;

    test_allocator_construct(&allocator);

    ern = VRPSX_CALL4(embedding, externalize_pid,
            expected,
            &allocator.interface,
            &outer_pid,
            &outer_size);

#if defined(__CYGWIN__)

    if (ern != VRPSX_CONSTANT(&embedding->system, enosys)) {
        printEError(HERE, embedding, "externalize_pid", ern);
        goto destroy_allocator;
    }
    if (outer_pid != NULL) {
        PRINT_HERE;
        printf("expected NULL result\n");
        goto destroy_allocator;
    }
    if (outer_size != 0) {
        PRINT_HERE;
        printf("expected empty result\n");
        goto destroy_allocator;
    }

#else /* not Cygwin */

    if (ern) {
        printEError(HERE, embedding, "externalize_pid", ern);
        goto destroy_allocator;
    }
    if (outer_size != sizeof(pid_t)) {
        PRINT_HERE;
        printf("outer_size %lu != %lu\n",
                (unsigned long)outer_size, (unsigned long)sizeof(pid_t));
        goto destroy_allocator;
    }
    if (*(pid_t*)outer_pid != expected) {
        PRINT_HERE;
        printf("pid %ld != pid %ld\n",
                (long)*(pid_t*)outer_pid, (long)expected);
        goto destroy_allocator;
    }
    if (assert_singular_allocation(HERE, &allocator, outer_pid, outer_size)) {
        goto destroy_allocator;
    }

#endif /* Cygwin? */

    if (outer_pid && outer_pid != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, outer_pid);
    }
    test_allocator_destroy(&allocator);
    return 0;

destroy_allocator:
    if (outer_pid && outer_pid != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, outer_pid);
    }
    test_allocator_destroy(&allocator);
    return -1;
}


static int assert_translation(
    where_t where,
    char const *expected,
    char const *actual,
    int (*comparison)(char const *, char const *))
{
    if (expected) {
        if (! actual) {
            print_where(where);
            printf("wrongly trivial translation\n");
            return -1;
        }
        if (comparison(actual, expected)) {
            print_where(where);
            printf("wrong translation: expected '%s' but saw '%s'\n",
                    expected, actual);
            return -1;
        }
    } else {
        if (actual) {
            print_where(where);
            printf("wrongly nontrivial translation: '%s'\n", actual);
            return -1;
        }
    }
    return 0;
}


int test_internalize_path(
    vrpsx_embedding_t *embedding)
{
    char *cleanup = NULL;
    test_allocator_t allocator;
    char const *outer_path;
    char const *expected_inner_path;
    char *inner_path = (char *)0xBAD;
    vrpsx_errno_t ern;

#if defined(__CYGWIN__)
    outer_path = cleanup = cygroot_relative("bin");
    expected_inner_path = "/usr/bin";
#else /* not Cygwin */
    outer_path = "/usr/bin";
    expected_inner_path = NULL;
#endif /* Cygwin? */

    test_allocator_construct(&allocator);

    ern = VRPSX_CALL3(embedding, internalize_path,
            outer_path,
            &allocator.interface,
            &inner_path);

    if (ern) {
        printEError(HERE, embedding, "internalize_path", ern);
        goto destroy_allocator;
    }

    if (assert_translation(HERE, expected_inner_path, inner_path, strcmp)) {
        goto destroy_allocator;
    }

    if (assert_singular_allocation(HERE, &allocator, inner_path,
                    (inner_path ? strlen(inner_path) + 1 : 0))) {
        goto destroy_allocator;
    }

    if (assert_nul_terminated_allocation(HERE, &allocator)) {
        goto destroy_allocator;
    }

    if (inner_path && inner_path != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, inner_path);
    }
    test_allocator_destroy(&allocator);
    free(cleanup);
    return 0;

destroy_allocator:
    if (inner_path && inner_path != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, inner_path);
    }
    test_allocator_destroy(&allocator);
    free(cleanup);
    return -1;
}


int test_externalize_path(
    vrpsx_embedding_t *embedding)
{
    char *cleanup = NULL;
    test_allocator_t allocator;
    char const *expected_outer_path;
    char *outer_path = (char *)0xBAD;
    vrpsx_errno_t ern;

#if defined(__CYGWIN__)
    expected_outer_path = cleanup = cygroot_relative("bin");
#else /* not Cygwin */
    expected_outer_path = NULL;
#endif /* Cygwin? */

    test_allocator_construct(&allocator);

    ern = VRPSX_CALL3(embedding, externalize_path,
            "/usr/bin",
            &allocator.interface,
            &outer_path);

    if (ern) {
        printEError(HERE, embedding, "externalize_path", ern);
        goto destroy_allocator;
    }

    if (assert_translation(HERE, expected_outer_path, outer_path, strcasecmp)){
        goto destroy_allocator;
    }

    if (assert_singular_allocation(HERE, &allocator, outer_path,
                    (outer_path ? strlen(outer_path) + 1 : 0))) {
        goto destroy_allocator;
    }

    if (assert_nul_terminated_allocation(HERE, &allocator)) {
        goto destroy_allocator;
    }

    if (outer_path && outer_path != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, outer_path);
    }
    test_allocator_destroy(&allocator);
    free(cleanup);
    return 0;

destroy_allocator:
    if (outer_path && outer_path != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, outer_path);
    }
    test_allocator_destroy(&allocator);
    free(cleanup);
    return -1;
}


int test_internalize_execution_path(
    vrpsx_embedding_t *embedding)
{
    char *cleanup[3] = { 0, 0, 0 };
    int cleanup_index;
    test_allocator_t allocator;
    char const *outer_path;
    char const *expected_inner_path;
    char *inner_path = (char *)0xBAD;
    vrpsx_errno_t ern;

    test_allocator_construct(&allocator);

#if defined(__CYGWIN__)
    cleanup[0] = cygroot_relative("home");
    cleanup[1] = cygroot_relative("bin");
    outer_path = cleanup[2] = execution_path_concat(cleanup[0], cleanup[1]);
    expected_inner_path = "/home:/usr/bin";
#else /* not Cygwin */
    outer_path = "/home:/usr/bin";
    expected_inner_path = NULL;
#endif /* Cygwin? */

    ern = VRPSX_CALL3(embedding, internalize_execution_path,
            outer_path,
            &allocator.interface,
            &inner_path);

    if (ern) {
        printEError(HERE, embedding, "internalize_execution_path", ern);
        goto destroy_allocator;
    }

    if (assert_translation(HERE, expected_inner_path, inner_path, strcmp)) {
        goto destroy_allocator;
    }

    if (assert_singular_allocation(HERE, &allocator, inner_path,
                    (inner_path ? strlen(inner_path) + 1 : 0))) {
        goto destroy_allocator;
    }

    if (assert_nul_terminated_allocation(HERE, &allocator)) {
        goto destroy_allocator;
    }

    outer_path = "";
    expected_inner_path = NULL;

    if (inner_path && inner_path != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, inner_path);
    }
    inner_path = (void *)0xBAD;

    ern = VRPSX_CALL3(embedding, internalize_execution_path,
            outer_path,
            &allocator.interface,
            &inner_path);

    if (ern) {
        printEError(HERE, embedding, "internalize_execution_path", ern);
        goto destroy_allocator;
    }

    if (assert_translation(HERE, expected_inner_path, inner_path, strcmp)) {
        goto destroy_allocator;
    }

    if (assert_singular_allocation(HERE, &allocator, inner_path,
                    (inner_path ? strlen(inner_path) + 1 : 0))) {
        goto destroy_allocator;
    }

    if (assert_nul_terminated_allocation(HERE, &allocator)) {
        goto destroy_allocator;
    }

    if (inner_path && inner_path != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, inner_path);
    }
    test_allocator_destroy(&allocator);
    for (cleanup_index = sizeof cleanup / sizeof *cleanup; cleanup_index--;) {
        free(cleanup[cleanup_index]);
    }
    return 0;

destroy_allocator:
    if (inner_path && inner_path != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, inner_path);
    }
    test_allocator_destroy(&allocator);
    for (cleanup_index = sizeof cleanup / sizeof *cleanup; cleanup_index--;) {
        free(cleanup[cleanup_index]);
    }
    return -1;
}


int test_externalize_execution_path(
    vrpsx_embedding_t *embedding)
{
    char *cleanup[3] = { 0, 0, 0 };
    int cleanup_index;
    test_allocator_t allocator;
    char const *expected_outer_path;
    char *outer_path = (char *)0xBAD;
    vrpsx_errno_t ern;

#if defined(__CYGWIN__)
    cleanup[0] = cygroot_relative("home");
    cleanup[1] = cygroot_relative("bin");
    expected_outer_path = cleanup[2]
            = execution_path_concat(cleanup[0], cleanup[1]);
#else /* not Cygwin */
    expected_outer_path = NULL;
#endif /* Cygwin? */

    test_allocator_construct(&allocator);

    ern = VRPSX_CALL3(embedding, externalize_execution_path,
            "/home:/usr/bin",
            &allocator.interface,
            &outer_path);

    if (ern) {
        printEError(HERE, embedding, "externalize_execution_path", ern);
        goto destroy_allocator;
    }

    if (assert_translation(HERE, expected_outer_path, outer_path, strcasecmp)){
        goto destroy_allocator;
    }

    if (assert_singular_allocation(HERE, &allocator, outer_path,
                    (outer_path ? strlen(outer_path) + 1 : 0))) {
        goto destroy_allocator;
    }

    if (assert_nul_terminated_allocation(HERE, &allocator)) {
        goto destroy_allocator;
    }

#if defined(__CYGWIN__)
    expected_outer_path = ".";
#else /* not Cygwin */
    expected_outer_path = NULL;
#endif /* Cygwin? */

    if (outer_path && outer_path != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, outer_path);
    }
    outer_path = (void *)0xBAD;

    ern = VRPSX_CALL3(embedding, externalize_execution_path,
            "",
            &allocator.interface,
            &outer_path);

    if (ern) {
        printEError(HERE, embedding, "externalize_execution_path", ern);
        goto destroy_allocator;
    }

    if (assert_translation(HERE, expected_outer_path, outer_path, strcasecmp)){
        goto destroy_allocator;
    }

    if (assert_singular_allocation(HERE, &allocator, outer_path,
                    (outer_path ? strlen(outer_path) + 1 : 0))) {
        goto destroy_allocator;
    }

    if (assert_nul_terminated_allocation(HERE, &allocator)) {
        goto destroy_allocator;
    }

    if (outer_path && outer_path != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, outer_path);
    }
    test_allocator_destroy(&allocator);
    for (cleanup_index = sizeof cleanup / sizeof *cleanup; cleanup_index--;) {
        free(cleanup[cleanup_index]);
    }
    return 0;

destroy_allocator:
    if (outer_path && outer_path != (void *)0xBAD) {
        VRPSX_CALL1(&allocator.interface, free_memory, outer_path);
    }
    test_allocator_destroy(&allocator);
    for (cleanup_index = sizeof cleanup / sizeof *cleanup; cleanup_index--;) {
        free(cleanup[cleanup_index]);
    }
    return -1;
}


typedef int test_t(vrpsx_embedding_t *);

typedef struct test_definition_struct {
    const char *name;
    test_t *function;
} test_definition_t;

#define TEST_DEFINITION(NAME)                                   \
    { VRPSX_STRINGIZE(NAME), & VRPSX_CAT2(test_, NAME) }

static test_definition_t const allTests[] = {
    TEST_DEFINITION(internalize_pid),
    TEST_DEFINITION(externalize_pid),
    TEST_DEFINITION(internalize_path),
    TEST_DEFINITION(externalize_path),
    TEST_DEFINITION(internalize_execution_path),
    TEST_DEFINITION(externalize_execution_path),
    { NULL, NULL }
};


VRPSX_DECLARE_IDENTIFIED_FUNCTION(static, main, vrpsx_embedding_main);

static int VRPSX_IDENTIFIED_FUNCTION_NAME(main, vrpsx_embedding_main)(
    vrpsx_embedding_initial_t *initial)
{
    vrpsx_embedding_t *embedding = initial->embedding;
    test_definition_t const *test;
    int exit_code = EXIT_SUCCESS;
    test_timer_t timer;
    struct timespec t0, t1;

#if defined(__CYGWIN__)
    cygroot_initialize();
#endif

    test_timer_construct(&timer);

    for (test = allTests; test->name; ++test) {
        printf("%s:\n", test->name);

        clock_gettime(CLOCK_REALTIME, &t0);

        if (test->function(embedding)) {
            exit_code = EXIT_FAILURE;
        }

        clock_gettime(CLOCK_REALTIME, &t1);

        test_timer_add_difference(&timer, test->name, &t0, &t1);
    }

    test_timer_report(&timer);

    test_timer_destroy(&timer);

    return exit_code;
}

vrpsx_identified_function_t const VRPSX_IDENTIFIED_FUNCTION_LIST[] = {
    VRPSX_MENTION_IDENTIFIED_FUNCTION(main, vrpsx_embedding_main),
    VRPSX_IDENTIFIED_FUNCTION_LIST_TERMINATOR
};
