/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Declares utilities to compute the "recursive digest"
 * of a header file as defined by "vrpsx_identified.h".
 */

#define _POSIX_C_SOURCE 200809L

#include "vrpsx_identify_lib.h"
#include <vrpsx_ostreambuf_default.h>
#include <vrpsx_md5.h>
#include <sys/stat.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>


#define MALLOC(TYPE)  ((TYPE *)malloc(sizeof(TYPE)))


void hash_map_init(
    hash_map_t *map)
{
    map->buckets = NULL;
    map->bucket_count = 0;
    map->entry_count = 0;
}

void hash_map_destroy(
    hash_map_t *map)
{
    hash_entry_t *next, *doomed;
    size_t index;

    for (index = 0; index < map->bucket_count; ++index) {
        next = map->buckets[index];
        while (next) {
            doomed = next;
            next = next->next;
            VRPSX_CALL0(doomed, disown);
        }
    }
}

int hash_map_insert(
    hash_map_t *map,
    hash_entry_t *newEnt,
    int *created,
    hash_entry_t **finalEnt)
{
    size_t index, new_index, new_count;
    hash_entry_t *current, *moved, **new_buckets;

    if (created) {
        *created = 0;
    }

    if (finalEnt) {
        *finalEnt = NULL;
    }

    index = 0;
    if (map->buckets) {
        index = newEnt->hash % map->bucket_count;
        for (current = map->buckets[index]; current; current = current->next) {
            if (VRPSX_CALL1(current, equal, newEnt)) {
                VRPSX_CALL0(newEnt, disown);
                if (finalEnt) {
                    *finalEnt = current;
                }
                return 0;
            }
        }
    }

    if (! map->buckets) {
        /* No bucket array at all; create one with a single bucket.
         */
        map->buckets = calloc(1, sizeof *map->buckets);
        if (! map->buckets) {
            VRPSX_CALL0(newEnt, disown);
            return errno;
        }
        map->bucket_count = 1;
    }

    /* Add new entry.
     */
    newEnt->next = map->buckets[index];
    map->buckets[index] = newEnt;
    ++ map->entry_count;

    if (created) {
        *created = 1;
    }

    if (finalEnt) {
        *finalEnt = newEnt;
    }

    if (map->entry_count <= (map->bucket_count << 1)) {
        /* Existing bucket array is still big enough.
         */
        return 0;
    }

    /* Buckets are getting overfull; expand the bucket array.
     */
    new_count = map->bucket_count << 1;
    if (new_count < (map->entry_count + 1) >> 1) {
        new_count = (map->entry_count + 1) >> 1;
    }
    new_buckets = (hash_entry_t **)calloc(new_count, sizeof *map->buckets);
    if (! new_buckets) {
        /* We have not actually failed, just grown more inefficient.
         */
        return 0;
    }

    /* Move entries over to the new bucket array.
     */
    for (index = 0; index < map->bucket_count; ++index) {
        current = map->buckets[index];
        while (current) {
            moved = current;
            current = current->next;

            new_index = moved->hash % new_count;
            moved->next = new_buckets[new_index];
            new_buckets[new_index] = moved;
        }
    }

    /* Free the old bucket array and update the map to use the new one.
     */
    free(map->buckets);
    map->buckets = new_buckets;
    map->bucket_count = new_count;
    return 0;
}

void hash_map_erase(
    hash_map_t *map,
    hash_entry_t *doomed)
{
    size_t index;
    hash_entry_t **next;

    assert (doomed);
    assert (map->buckets);

    index = doomed->hash % map->bucket_count;
    for (next = &map->buckets[index]; *next; next = &(*next)->next) {
        if (*next == doomed) {
            break;
        }
    }

    assert (*next == doomed);
    *next = doomed->next;
    VRPSX_CALL0(doomed, disown);
}


static hash_entry_vtbl_t const recursive_digest_entry_vtbl;

/* Creates an instance of recursive_digest_entry_t for the specified header.
 * Returns an errno-style error number.
 */
static int recursive_digest_entry_create(
    dev_t dev,                             /* Device containing header. */
    ino_t ino,                             /* Header file inode. */
    recursive_digest_entry_t **ent)        /* Overwritten with pointer to
                                            * new entry, or NULL on failure. */
{
    recursive_digest_entry_t *obj;

    *ent = NULL;

    obj = MALLOC(recursive_digest_entry_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &recursive_digest_entry_vtbl;
    obj->interface.next = NULL;
    obj->interface.hash = (size_t)dev * 5 + (size_t)ino;
    obj->dev = dev;
    obj->ino = ino;
    memset(obj->digest, 0, sizeof obj->digest);

    *ent = obj;
    return 0;
}

static void recursive_digest_entry_disown(
    hash_entry_t *ent)
{
    VRPSX_UPCAST_NONNULL_DECL(
        recursive_digest_entry_t, , obj, interface, ent);
    free(obj);
}

static int recursive_digest_entry_equal(
    hash_entry_t *ent,
    hash_entry_t *other)
{
    VRPSX_UPCAST_NONNULL_DECL(
        recursive_digest_entry_t, , obj, interface, ent);
    VRPSX_UPCAST_NONNULL_DECL(
        recursive_digest_entry_t, , other_obj, interface, other);

    assert (other->vtbl == &recursive_digest_entry_vtbl);

    return obj->ino == other_obj->ino
        && obj->dev == other_obj->dev;
}

static hash_entry_end_t const hash_entry_end;

static hash_entry_vtbl_t const recursive_digest_entry_vtbl = {
    &recursive_digest_entry_disown,
    &recursive_digest_entry_equal,
    &hash_entry_end
};


void recursive_digest_map_init(
    recursive_digest_map_t *map)
{
    hash_map_init(&map->generic);
}

void recursive_digest_map_destroy(
    recursive_digest_map_t *map)
{
    hash_map_destroy(&map->generic);
}

int recursive_digest_map_insert(
    recursive_digest_map_t *map,
    dev_t dev,
    ino_t ino,
    int *created,
    recursive_digest_entry_t **finalEnt)
{
    recursive_digest_entry_t *myNewEnt;
    hash_entry_t *myFinalEnt;

    if (created) {
        *created = 0;
    }

    if (finalEnt) {
        *finalEnt = NULL;
    }

    int ern = recursive_digest_entry_create(dev, ino, &myNewEnt);
    if (ern) {
        return ern;
    }

    ern = hash_map_insert(
        &map->generic, &myNewEnt->interface, created, &myFinalEnt);
    if (ern) {
        return ern;
    }

    if (finalEnt) {
        *finalEnt = VRPSX_UPCAST_NONNULL(
            recursive_digest_entry_t, interface, myFinalEnt);
    }

    return 0;
}

void recursive_digest_map_erase(
    recursive_digest_map_t *map,
    recursive_digest_entry_t *doomed)
{
    hash_map_erase(&map->generic, &doomed->interface);
}


size_t quoted_length_of_include(
    char const *qpath)
{
    char const *end;
    char closer;

    switch (*qpath) {
        case '"': {
            closer = '"';
            break;
        }
        case '<': {
            closer = '>';
            break;
        }
        default: {
            assert (0);
            closer = '\0'; /* avoid compiler warning */
            break;
        }
    }

    for (end = qpath + 1; *end++ != closer;) {}
    return end - qpath;
}


/* If error_message != NULL, then stores in *error_message an error message
 * based on the given error number, which is assumed to result from some
 * general operation not specifically related to a particular header or file,
 * such as a memory allocation error.
 *
 * On failure--memory allocation failure--stores NULL in *error_message.
 */
static void compose_general_error(
    char **error_message,
    int ern)
{
    if (! error_message) {
        return;
    }

    // We are not sure of the standardization of "strdup";
    // therefore we have this somewhat more verbose code:

    const char *temporary = strerror(ern);
    size_t size = strlen(temporary) + 1;
    *error_message = malloc(size);
    memcpy(*error_message, temporary, size);
}

/* If error_message != NULL, then stores in *error_message an error message
 * based on he given error number, which is assumed to result from a failure
 * to write to the debug output stream.
 *
 * On failure--memory allocation failure--stores NULL in *error_message.
 */
static void compose_debug_error(
    char **error_message,
    int ern)
{
    static char const prefix[] = "debug output stream: ";
    size_t block1, block2;
    char *next;

    if (! error_message) {
        return;
    }

    block1 = (sizeof prefix) - 1;
    block2 = strlen(strerror(ern)) + 1;

    *error_message = (char *)malloc(block1 + block2);
    if (! *error_message) {
        return;
    }

    next = *error_message;
    MEMCPY_END(next, prefix, block1);
    memcpy(next, strerror(ern), block2);
}

/* If error_message != NULL, then stores in *error_message an
 * error message based on the given pathname and error number.
 *
 * On failure--memory allocation failure--stores NULL in *error_message.
 */
static void compose_path_error(
    char **error_message,
    char const *path,
    int ern)
{
    size_t block1, block3;
    char *next;

    if (! error_message) {
        return;
    }

    block1 = strlen(path);
    block3 = strlen(strerror(ern)) + 1;

    *error_message = (char *)malloc(block1 + 2 + block3);
    if (! *error_message) {
        return;
    }

    next = *error_message;
    MEMCPY_END(next, path, block1);
    MEMCPY_END(next, ": ", 2);
    memcpy(next, strerror(ern), block3);
}

/* If error_message != NULL, then stores in *error_message an error message
 * based on the given quoted inclusion pathname and error number.
 *
 * On failure--memory allocation failure--stores NULL in *error_message.
 */
static void compose_qpath_error(
    char **error_message,
    char const *context,          /* If non-NULL, the including header. */
    long context_line,            /* If context != NULL, the line number. */
    char const *wd,               /* Working dir for "..." includes. */
    char const *qpath,            /* See quoted_length_of_include. */
    int ern)
{
    char linebuf[1 + 1 + (sizeof context_line * CHAR_BIT + 2) / 3 + 1 + 1];
    size_t block1, block2, block3, block5, block7;
    char *next;

    if (! error_message) {
        return;
    }

    if (context) {
        sprintf(linebuf, ":%ld:", context_line);
        block1 = strlen(context);
        block2 = strlen(linebuf);
    } else {
        block1 = 0;
        block2 = 0;
    }

    block3 = strlen(wd);
    block5 = quoted_length_of_include(qpath);
    block7 = strlen(strerror(ern)) + 1;

    *error_message = (char *)malloc(
        block1 + block2 + block3 + 2 + block5 + 2 + block7);
    if (! *error_message) {
        return;
    }

    next = *error_message;
    if (context) {
        MEMCPY_END(next, context, block1);
        MEMCPY_END(next, linebuf, block2);
    }
    MEMCPY_END(next, wd, block3);
    MEMCPY_END(next, ": ", 2);
    MEMCPY_END(next, qpath, block5);
    MEMCPY_END(next, ": ", 2);
    memcpy(next, strerror(ern), block7);
}


/* The state of the transformation from header text into MD5 input.
 *
 * Output errors are available from sink->error; all other errors
 * are reported as ESYNTAX, and are not stored in this struct.
 */
typedef struct recursive_digest_struct {
    char **error_message;
    recursive_digest_map_t *map;
    char const *const *include_directories;
    char const *header;
    char const *header_directory;
    char const *current;
    char const *end;
    vrpsx_ostreambuf_t *sink;
    int postcr;
    long line;
} recursive_digest_t;


#define IS_HSPACE(CH)  ((CH) == ' ' || (CH) == '\t')
#define IS_VSPACE(CH)  ((CH) == '\n' || (CH) == '\r')
#define IS_WSPACE(CH)  (IS_HSPACE(CH) || IS_VSPACE(CH))

#define ADVANCE1(STATE)  {                                              \
    recursive_digest_t *state_b089bf1e = (STATE);                       \
                                                                        \
    if (state_b089bf1e->current[0] == '\n'                              \
        || (state_b089bf1e->current[0] == '\r'                          \
            && (state_b089bf1e->current + 1 >= state_b089bf1e->end      \
                || state_b089bf1e->current[1] != '\n'))) {              \
        ++ state_b089bf1e->line;                                        \
    }                                                                   \
                                                                        \
    ++ state_b089bf1e->current;                                         \
}

static void advancen(
    recursive_digest_t *state,
    size_t count)
{
    while (count--) {
        ADVANCE1(state);
    }
}


/* Outside of string/character literals and comments,
 * skips horizontal whitespace.
 */
static void skip_horizontal_space(
    recursive_digest_t *state)
{
    while (state->current != state->end && IS_HSPACE(*state->current)) {
        ++ state->current;
    }
}

/* Like skip_horizontal_space(), but if there was horizontal whitespace,
 * and it was not at the end of a line/header, then also also writes ' '
 * to state->sink.
 */
static void compress_horizontal_space(
    recursive_digest_t *state)
{
    int found = (state->current != state->end && IS_HSPACE(*state->current));

    skip_horizontal_space(state);

    if (found && state->current != state->end && ! IS_VSPACE(*state->current)){
        VRPSX_OSTREAMBUF_PUT1(state->sink, ' ', return);
    }
}

/* Outside of string/character literals and comments, skip whitespace.
 */
static void skip_space(
    recursive_digest_t *state)
{
    while (state->current != state->end && IS_WSPACE(*state->current)) {
        ADVANCE1(state);
    }
}


/* Copies a single character to state->sink, as-is.
 */
static void output1(
    recursive_digest_t *state)
{
    VRPSX_OSTREAMBUF_PUT1(state->sink, *state->current, return);
    ADVANCE1(state);
}

/* Copies several characters to state->sink, as-is.
 */
static void outputn(
    recursive_digest_t *state,
    size_t count)
{
    VRPSX_OSTREAMBUF_PUTN(state->sink, state->current, count, return);
    advancen(state, count);
}


static char const xdigit[16] = "0123456789abcdef";

/* Assumes we are at the start of line that contains a non-whitespace
 * character, after initial whitespace is removed, and if the line is
 * an include directive, then transforms it into MD5 input written to
 * state->sink.  Otherwise may correctly transform some amount of text.
 *
 * EXCEPT: The portion of the line containing the include directive
 * that comes after the include pathname may be left unprocessed;
 * the caller can apply general rules to that portion of the line.
 */
static int transform_include_directive(
    char **error_message,
    recursive_digest_t *state)
{
    static char const EXCLUDE[] = "EXCLUDE_FROM_INTERFACE_IDENTIFIER";

    char const *closer;
    char const *lineEnd;
    char const *candidateEnd;
    char const *exclusion;
    char closerValue;
    size_t octetIndex;
    int ern;
    vrpsx_uint8_t digest[MD5_DIGEST_LENGTH];

    if (*state->current != '#') {
        /* Not a preprocessor directive.
         */
        return state->sink->error;
    }

    output1(state);

    /* The definition of "recursive digest" says: "Delete any ' ' following
     * a '#' at the start of a line, except inside of C string/character
     * literals and comments.  start of a line out."  We do that here:
     */
    skip_horizontal_space(state);

    if (state->end - state->current < 7
            || memcmp(state->current, "include", 7)) {
        /* Not an include directive.
         */
        return state->sink->error;
    }

    outputn(state, 7);
    if (state->sink->error) {
        return state->sink->error;
    }

    compress_horizontal_space(state);
    if (state->sink->error) {
        return state->sink->error;
    }

    if (state->current == state->end) {
        /* Missing pathname of file to include.
         */
        return state->sink->error;
    }

    switch (*state->current) {
        case '<':
            closerValue = '>';
            break;
        case '"':
            closerValue = '"';
            break;
        default: {
            /* Wrong kind of pathname quotation.
             */
            return state->sink->error;
        }
    }

    /* Find end of include pathname.
     */
    for (closer = state->current + 1;
         closer != state->end
                 && ! IS_VSPACE(*closer)
                 && *closer != closerValue;
         ++closer) {}
    if (closer == state->end || IS_VSPACE(*closer)) {
        /* Unterminated include pathname.
         */
        outputn(state, state->end - state->current);
        return state->sink->error;
    }

    /* Find end of line containing include directive.
     */
    lineEnd = closer;
    while (lineEnd != state->end && ! IS_VSPACE(*lineEnd)) {
        ++lineEnd;
    }

    /* Check for exclusion.  A naive algorithm suffices not
     * just because this is not a performance-critical case
     * but also because the 'X' early in the search string
     * limits actual time spent on overlapping comparisons.
     */
    exclusion = candidateEnd = NULL;
    if (sizeof EXCLUDE - 2 < lineEnd - closer) {
        candidateEnd = lineEnd - (sizeof EXCLUDE - 2);
        for (exclusion = closer; exclusion != candidateEnd; ++exclusion) {
            if (0 == memcmp(exclusion, EXCLUDE, sizeof EXCLUDE - 1)) {
                break;
            }
        }
    }

    if (exclusion != candidateEnd) {
        /* Excluded.  Emit the quoted pathname as-is.
         */
        outputn(state, closer + 1 - state->current);
    } else {
        /* Not excluded, so recursively process the include file.
         * Debug output streams do NOT propagate to recursive calls.
         */
        ern = recursive_digest(
            error_message,
            state->map,
            state->include_directories,
            state->header,
            state->line,
            state->header_directory,
            state->current,
            digest,
            NULL);
        if (ern) {
            return ern;
        }

        /* The definition of "recursive digest" says we should replace
         * the pathname of the inclusion with the 32-digit big-endian
         * lowercase hexadecimal form of its recursive digest.
         */
        for (octetIndex = 0; octetIndex < MD5_DIGEST_LENGTH; ++octetIndex) {
            VRPSX_OSTREAMBUF_PUT1(state->sink, xdigit[digest[octetIndex] >> 4],
                    return state->sink->error);
            VRPSX_OSTREAMBUF_PUT1(state->sink, xdigit[digest[octetIndex] & 0xF],
                    return state->sink->error);
        }

        state->current = closer + 1;
    }

    return state->sink->error;
}


/* If we are at the start of a C99 "//" comment, then transform it;
 * otherwise do nothing.
 */
static void transform_c99_comment(
    recursive_digest_t *state)
{
    if (state->end - state->current < 2
            || state->current[0] != '/'
            || state->current[1] != '/') {
        return;
    }

    output1(state);
    output1(state);

    while (state->current != state->end
            && ! IS_VSPACE(*state->current)
            && ! state->sink->error) {
        if (IS_HSPACE(*state->current)) {
            compress_horizontal_space(state);
        } else {
            output1(state);
        }
    }
}


/* If we are at the start of a C character/string literal, outputs it as-is;
 * otherwise may transform some part of the text that resembles the initial
 * portion of a C character/string literal.
 */
static void output_text_literal(
    recursive_digest_t *state)
{
    int escaped = 0;
    char quote;

    if (state->current == state->end) {
        return;
    }

    quote = *state->current;
    if (quote != '"' && quote != '\'') {
        return;
    }

    output1(state);

    while (state->current != state->end
            && ! IS_VSPACE(*state->current)
            && ! state->sink->error) {
        if (*state->current == quote) {
            if (! escaped) {
                output1(state);
                return;
            }
            escaped = 0;
            output1(state);
        } else if (*state->current == '\\') {
            escaped = ! escaped;
            output1(state);
        } else {
            escaped = 0;
            output1(state);
        }
    }
}


/* Assumes we are at the start of line that contains a non-whitespace
 * character, after initial whitespace is removed, and transforms that line,
 * not including any terminating newline.
 */
static int transform_line(
    char **error_message,
    recursive_digest_t *state,
    int *in_c89_comment)  /* Input/Output: are we inside a C89 comment? */
{
    int ern;

    if (! *in_c89_comment) {
        ern = transform_include_directive(error_message, state);
        if (ern) {
            return ern;
        }
    }

    while (state->current != state->end && ! state->sink->error) {
        switch (*state->current) {
            case '\n':
            case '\r': {
                return state->sink->error;
            }
            case ' ':
            case '\t': {
                compress_horizontal_space(state);
                break;
            }
            case '*': {
                if (*in_c89_comment
                        && state->current + 1 != state->end
                        && state->current[1] == '/') {
                    *in_c89_comment = 0;
                    output1(state);
                }
                output1(state);
                break;
            }
            case '/': {
                if (*in_c89_comment || state->current + 1 == state->end) {
                    output1(state);
                } else if (state->current[1] == '*') {
                    *in_c89_comment = 1;
                    output1(state);
                    output1(state);
                } else if (state->current[1] == '/') {
                    transform_c99_comment(state);
                } else {
                    output1(state);
                }
                break;
            }
            case '\'':
            case '"': {
                if (*in_c89_comment) {
                    output1(state);
                } else {
                    output_text_literal(state);
                }
                break;
            }
            default: {
                output1(state);
                break;
            }
        }
    }

    return state->sink->error;
}


/* Similar to recursive_digest(), but requires the caller to have
 * already resolved the header location, loaded it into memory,
 * excluded the case in which the digest is already computed, created
 * a new blank map entry for the digest, and set up an abstract stream
 * that receives characters to be digested and possibly also writes
 * debug output.
 */
static int recursive_digest_in_memory(
    char **error_message,
    recursive_digest_map_t *map,
    char const *const *include_directories,
    char const *header,
    char const *header_directory,
    char *begin,
    char *end,
    vrpsx_ostreambuf_t *sink)
{
    recursive_digest_t state;
    int in_c89_comment;
    int ern;

    if (error_message) {
        *error_message = NULL;
    }

    if (begin == end) {
        return 0;
    }

    state.error_message = error_message;
    state.map = map;
    state.include_directories = include_directories;
    state.header = header;
    state.header_directory = header_directory;
    state.current = begin;
    state.end = end;
    state.sink = sink;
    state.line = 1;

    in_c89_comment = 0;

    skip_space(&state);

    while (state.current != state.end) {
        ern = transform_line(error_message, &state, &in_c89_comment);
        if (ern) {
            if (sink->error && error_message && ! *error_message) {
                compose_debug_error(error_message, sink->error);
            }
            return ern;
        }

        skip_space(&state);
        VRPSX_OSTREAMBUF_PUT1(sink, '\n', return sink->error);
    }

    if (sink->error && error_message && ! *error_message) {
        compose_debug_error(error_message, sink->error);
    }
    return sink->error;
}


typedef struct tee_ostreambuf_struct {
    vrpsx_ostreambuf_t interface;
    vrpsx_ostreambuf_t *delegate[2];  // NON-OWNING
    char block[64];
} tee_ostreambuf_t;

static vrpsx_ostreambuf_vtbl_t const tee_ostreambuf_vtbl;

static int tee_ostreambuf_create(
    vrpsx_ostreambuf_t *first,
    vrpsx_ostreambuf_t *second,
    vrpsx_ostreambuf_t **osb)
{
    *osb = NULL;

    tee_ostreambuf_t *obj = MALLOC(tee_ostreambuf_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &tee_ostreambuf_vtbl;
    obj->interface.current = obj->block;
    obj->interface.end = obj->block + sizeof obj->block;
    obj->interface.error = 0;
    obj->delegate[0] = first;
    obj->delegate[1] = second;

    *osb = &obj->interface;
    return 0;
}

static int tee_ostreambuf_disown(
    vrpsx_ostreambuf_t **osb)
{
    vrpsx_ostreambuf_t *doomed = *osb;
    *osb = NULL;
    int ern = VRPSX_CALL0(doomed, flush);
    VRPSX_UPCAST_NONNULL_DECL(tee_ostreambuf_t, , obj, interface, doomed);
    free(obj);
    return ern;
}

static int tee_ostreambuf_flush(
    vrpsx_ostreambuf_t *osb)
{
    VRPSX_UPCAST_NONNULL_DECL(
        tee_ostreambuf_t, , obj, interface, osb);
    int index;

    if (obj->interface.error) {
        return obj->interface.error;
    }

    for (index = 0; index < 2; ++index) {
        VRPSX_OSTREAMBUF_PUTN(
            obj->delegate[index],
            obj->block,
            obj->interface.current - obj->block,
            {
                obj->interface.error = obj->delegate[index]->error;
                return obj->interface.error;
            });
    }

    obj->interface.current = obj->block;
    return 0;
}

static int tee_ostreambuf_putn(
    vrpsx_ostreambuf_t *osb,
    char const *begin,
    size_t size)
{
    VRPSX_UPCAST_NONNULL_DECL(
        tee_ostreambuf_t, , obj, interface, osb);
    int index;

    if (obj->interface.error) {
        return obj->interface.error;
    }

    if (VRPSX_CALL0(&obj->interface, flush)) {
        return obj->interface.error;
    }

    for (index = 0; index < 2; ++index) {
        VRPSX_OSTREAMBUF_PUTN(
            obj->delegate[index],
            begin,
            size,
            {
                obj->interface.error = obj->delegate[index]->error;
                return obj->interface.error;
            });
    }

    return 0;
}

static vrpsx_ostreambuf_end_t const vrpsx_ostreambuf_end;

static vrpsx_ostreambuf_vtbl_t const tee_ostreambuf_vtbl = {
    &tee_ostreambuf_disown,
    &tee_ostreambuf_flush,
    &vrpsx_ostreambuf_put1_from_flush,
    &tee_ostreambuf_putn,
    &vrpsx_ostreambuf_end
};


typedef struct md5_ostreambuf_struct {
    vrpsx_ostreambuf_t interface;
    MD5_CTX ctx;
    vrpsx_uint8_t block[MD5_BLOCK_LENGTH];
} md5_ostreambuf_t;

static vrpsx_ostreambuf_vtbl_t const md5_ostreambuf_vtbl;

static int md5_ostreambuf_create(
    md5_ostreambuf_t **md5)
{
    *md5 = NULL;

    md5_ostreambuf_t *obj = MALLOC(md5_ostreambuf_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &md5_ostreambuf_vtbl;
    obj->interface.current = (char *)obj->block;
    obj->interface.end = (char *)obj->block + sizeof obj->block;
    obj->interface.error = 0; /* And it will stay 0: no errors are possible. */
    MD5Init(&obj->ctx);

    *md5 = obj;
    return 0;
}

static void md5_ostreambuf_final(
    md5_ostreambuf_t *obj,
    vrpsx_uint8_t digest[MD5_DIGEST_LENGTH])
{
    MD5Update(&obj->ctx, obj->block,
            obj->interface.current - (char *)obj->block);
    obj->interface.current = (char *)obj->block;

    MD5Final(digest, &obj->ctx);
}

static int md5_ostreambuf_disown(
    vrpsx_ostreambuf_t **osb)
{
    vrpsx_ostreambuf_t *doomed = *osb;
    *osb = NULL;
    int ern = VRPSX_CALL0(doomed, flush);
    VRPSX_UPCAST_NONNULL_DECL(md5_ostreambuf_t, , obj, interface, doomed);
    free(obj);
    return ern;
}

static int md5_ostreambuf_flush(
    vrpsx_ostreambuf_t *osb)
{
    /* There is no point in attempting to empty the buffer because
     * until the data are complete the digest cannot be finalized.
     */
    return 0;
}

static int md5_ostreambuf_put1(
    vrpsx_ostreambuf_t *osb,
    char ch)
{
    VRPSX_UPCAST_NONNULL_DECL(md5_ostreambuf_t, , obj, interface, osb);

    if (obj->interface.current == (char *)obj->block + sizeof obj->block) {
        MD5Update(&obj->ctx, obj->block, sizeof obj->block);
        obj->interface.current = (char *)obj->block;
    }

    *obj->interface.current ++ = (vrpsx_uint8_t)ch;
    return 0;
}

static vrpsx_ostreambuf_vtbl_t const md5_ostreambuf_vtbl = {
    &md5_ostreambuf_disown,
    &md5_ostreambuf_flush,
    &md5_ostreambuf_put1,
    &vrpsx_ostreambuf_putn_from_put1,
    &vrpsx_ostreambuf_end
};


/* Locates a header.
 *
 * If error_message != NULL then this function sets *error_message to NULL
 * unless it successfully composed an error message for the failure, in which
 * case it stores that message, which the caller must deallocate with "free".
 * Composition fails only due to insufficient memory.
 *
 * The other output arguments are set to NULL on failure.
 */
static int resolve_include(
    char **error_message,
    char const *const *include_directories, /* NULL-terminated list of dirs. */
    char const *context,                  /* If non-NULL, including header.*/
    long context_line,                    /* If context != NULL, line number.*/
    char const *wd,                       /* Working dir for "..." includes. */
    char const *qpath,                    /* See quoted_length_of_include. */
    char **resolution,                    /* Output: resolved header pathname;
                                           * eventually pass to "free". */
    FILE **fp)                            /* Output: file handle for header;
                                           * eventually pass to "fclose". */
{
    char *next;
    char const *wd_list[2] = { (qpath[0] == '"' ? wd : NULL), NULL };
    char const *const *list_list[] = { wd_list, include_directories };
    char const *const *const *plist;
    char const *const *pdir;
    char const *dir;
    size_t qpath_length;
    size_t dir_length;
    int ern;

    if (error_message) {
        *error_message = NULL;
    }

    *resolution = NULL;
    *fp = NULL;

    qpath_length = quoted_length_of_include(qpath);

    /* If the header pathname is absolute, then just try it as-is:
     */
    if (qpath[1] == '/') {
        *resolution = (char *)malloc(qpath_length - 2 + 1);
        if (! *resolution) {
            ern = errno;
            compose_general_error(error_message, ern);
            return ern;
        }
        next = *resolution;
        MEMCPY_END(next, qpath + 1, qpath_length - 2);
        *next = '\0';

        *fp = fopen(*resolution, "r");
        if (! *fp) {
            ern = errno;
            compose_path_error(error_message, *resolution, ern);
            free(*resolution);
            *resolution = NULL;
            return ern;
        }

        return 0;
    }

    /* Header pathname is relative, so search the include directories.
     * For "..." includes, we start with the given working directory.
     */
    ern = 0;

    for (plist = list_list; ! *fp && plist != list_list + 2; ++plist) {
        for (pdir = *plist; ! *fp && *pdir; ++pdir) {
            dir = *pdir;

            /* First combine the include directory with the relative path:
             */
            dir_length = strlen(dir);
            while (dir_length >= 2 && dir[dir_length - 1] == '/') {
                -- dir_length;
            }
            *resolution = (char*)malloc(dir_length + 1 + qpath_length - 2 + 1);
            if (! *resolution) {
                ern = errno;
                free(*resolution);
                compose_general_error(error_message, ern);
                return ern;
            }
            next = *resolution;
            MEMCPY_END(next, dir, dir_length);
            *next++ = '/';
            MEMCPY_END(next, qpath + 1, qpath_length - 2);
            *next = '\0';

            /* Then attempt to open that candidate for the resolved path:
             */
            *fp = fopen(*resolution, "r");
            if (! *fp) {
                /* Currently, the first error determines the error number.
                 */
                if (! ern) {
                    ern = errno;
                }
                free(*resolution);
                *resolution = NULL;
            }
        }
    }

    if (! *fp) {
        /* Resolution failed in all directories.
         */
        if (! ern) {
            /* Should happen only if no directories were searched.
             */
            ern = ENOENT;
        }
        compose_qpath_error(
            error_message, context, context_line, wd, qpath, ern);
        return ern;
    }

    return 0;
}

/* Reads the given header file.
 * Arranges that [*begin, *end) is the in-memory copy.
 * You should eventually call free(*begin).
 * On failure sets *begin and *end to NULL.
 */
static int read_header(
    char **error_message,
    char const *path,
    FILE *fp,
    char **begin,
    char **end)
{
    long length;
    int ern;

    if (error_message) {
        *error_message = NULL;
    }

    *begin = NULL;
    *end = NULL;

    /* Compute the length of the header:
     */
    if (fseek(fp, 0, SEEK_END)) {
        ern = errno;
        fclose(fp);
        compose_path_error(error_message, path, ern);
        return ern;
    }
    length = ftell(fp);
    if (length > (size_t)-1) {
        ern = ENOMEM;
        fclose(fp);
        compose_path_error(error_message, path, ern);
        return ern;
    }

    /* Get back to the beginning of the header:
     */
    if (fseek(fp, 0, SEEK_SET)) {
        ern = errno;
        fclose(fp);
        compose_path_error(error_message, path, ern);
        return ern;
    }

    /* Read the entire header into memory:
     */
    *begin = (char *)malloc(length);
    if (! *begin) {
        ern = errno;
        fclose(fp);
        compose_general_error(error_message, ern);
        return ern;
    }
    if (fread(*begin, 1, (size_t)length, fp) != (size_t)length) {
        ern = errno;
        free(*begin);
        *begin = NULL;
        fclose(fp);
        compose_path_error(error_message, path, ern);
        return ern;
    }

    *end = *begin + length;
    return 0;
}

int recursive_digest(
    char **error_message,
    recursive_digest_map_t *map,
    char const *const *include_directories,
    char const *context,
    long context_line,
    char const *wd,
    char const *qpath,
    vrpsx_uint8_t digest[MD5_DIGEST_LENGTH],
    vrpsx_ostreambuf_t *debug)
{
    char *resolution = NULL;
    char *dir = NULL;
    FILE *fp = NULL;
    recursive_digest_entry_t *ent = NULL;
    char *begin = NULL;
    char *end = NULL;
    int created = 0;
    md5_ostreambuf_t *md5 = NULL;
    vrpsx_ostreambuf_t *tee = NULL;
    vrpsx_ostreambuf_t *sink;
    struct stat stat_buffer;
    size_t dir_length;
    int ern;

    if (error_message) {
        *error_message = NULL;
    }

    memset(digest, 0, MD5_DIGEST_LENGTH);

    ern = resolve_include(
        error_message,
        include_directories,
        context,
        context_line,
        wd,
        qpath,
        &resolution,
        &fp);
    if (ern) {
        return ern;
    }

    if (fstat(fileno(fp), &stat_buffer)) {
        ern = errno;
        fclose(fp);
        compose_path_error(error_message, resolution, ern);
        free(resolution);
        return ern;
    }

    ern = recursive_digest_map_insert(
        map, stat_buffer.st_dev, stat_buffer.st_ino, &created, &ent);
    if (ern) {
        fclose(fp);
        free(resolution);
        compose_general_error(error_message, ern);
        return ern;
    }

    if (! created) {
        /* At this point "digest" contains all zeros.
         */
        if (! memcmp(digest, ent->digest, MD5_DIGEST_LENGTH)) {
            /* We are currently computing this recursive digest, so
             * there must be an inclusion cycle that was not excluded.
             */
            fclose(fp);
            compose_path_error(error_message, resolution, ELOOP);
            free(resolution);
            return ELOOP;
        }

        /* Reuse prior result.
         */
        memcpy(digest, ent->digest, MD5_DIGEST_LENGTH);
        return 0;
    }

    ern = read_header(error_message, resolution, fp, &begin, &end);
    if (ern) {
        recursive_digest_map_erase(map, ent);
        fclose(fp);
        free(resolution);
        return ern;
    }

    if (fclose(fp)) {
        ern = errno;
        free(begin);
        begin = NULL;
        end = NULL;
        recursive_digest_map_erase(map, ent);
        compose_path_error(error_message, resolution, ern);
        free(resolution);
        return ern;
    }
    fp = NULL;

    /* Get the directory containing current header.
     *
     * Possible Limitation: We have not yet bothered to check whether
     * a C preprocessor would first follow symlinks before computing
     * the directory to use for relative header inclusions.
     */
    dir_length = strlen(resolution);
    while (dir_length && resolution[dir_length - 1] == '/') {
        --dir_length;
    }
    while (dir_length && resolution[dir_length - 1] != '/') {
        --dir_length;
    }
    while (dir_length && resolution[dir_length - 1] == '/') {
        --dir_length;
    }
    if (dir_length == 0) {
        dir_length = 1;
        dir = malloc(2);
        if (resolution[0] == '/') {
            dir[0] = '/';
        } else {
            dir[0] = '.';
        }
    } else {
        dir = malloc(dir_length + 1);
        memcpy(dir, resolution, dir_length);
    }
    dir[dir_length] = '\0';

    ern = md5_ostreambuf_create(&md5);
    if (ern) {
        free(begin);
        begin = NULL;
        end = NULL;
        recursive_digest_map_erase(map, ent);
        compose_path_error(error_message, resolution, ern);
        free(dir);
        free(resolution);
        return ern;
    }

    sink = &md5->interface;

    if (debug) {
        ern = tee_ostreambuf_create(sink, debug, &tee);
        if (ern) {
            sink = NULL;
            VRPSX_SINGLE_OWNER_CLEAR_IMPLEMENTATION(
                md5_ostreambuf_t, interface, vrpsx_ostreambuf_t, &md5);
            free(begin);
            begin = NULL;
            end = NULL;
            recursive_digest_map_erase(map, ent);
            compose_path_error(error_message, resolution, ern);
            free(dir);
            free(resolution);
            return ern;
        }
        sink = tee;
    }

    ern = recursive_digest_in_memory(
        error_message,
        map,
        include_directories,
        resolution,
        dir,
        begin,
        end,
        sink);
    free(begin);
    begin = NULL;
    end = NULL;
    free(dir);
    free(resolution);
    resolution = NULL;
    if (ern) {
        sink = NULL;
        VRPSX_SINGLE_OWNER_CLEAR(vrpsx_ostreambuf_t, &tee);
        VRPSX_SINGLE_OWNER_CLEAR_IMPLEMENTATION(
            md5_ostreambuf_t, interface, vrpsx_ostreambuf_t, &md5);
        recursive_digest_map_erase(map, ent);
        return ern;
    }
    sink = NULL;

    ern = 0;
    VRPSX_SINGLE_OWNER_CLOSE(vrpsx_ostreambuf_t, &tee, &ern);
    if (ern) {
        VRPSX_SINGLE_OWNER_CLEAR_IMPLEMENTATION(
            md5_ostreambuf_t, interface, vrpsx_ostreambuf_t, &md5);
        recursive_digest_map_erase(map, ent);
        compose_debug_error(error_message, ern);
        return ern;
    }

    md5_ostreambuf_final(md5, ent->digest);

    ern = 0;
    VRPSX_SINGLE_OWNER_CLOSE_IMPLEMENTATION(
        md5_ostreambuf_t, interface, vrpsx_ostreambuf_t, &md5, &ern);
    if (ern) {
        compose_debug_error(error_message, ern);
        return ern;
    }

    memcpy(digest, ent->digest, MD5_DIGEST_LENGTH);
    return 0;
}


static vrpsx_uint32_t recursive_digest_to_interface_identifier(
    vrpsx_uint8_t const digest[MD5_DIGEST_LENGTH], /* Recursive digest. */
    char const *name)                              /* Name within header. */
{
    MD5_CTX ctx;
    vrpsx_uint8_t const *dp;
    vrpsx_uint8_t const *ep;
    char digest_text[2 * MD5_DIGEST_LENGTH + 1], *cp;
    vrpsx_uint8_t pair_digest[MD5_DIGEST_LENGTH];

    /* Convert recursive digest to 32-digit big-endian lowercase hexadecimal.
     */
    for (cp = digest_text, dp = digest, ep = digest + MD5_DIGEST_LENGTH;
         dp != ep; cp += 2, ++dp) {
        cp[0] = xdigit[*dp >> 4];
        cp[1] = xdigit[*dp & 0xF];
    }

    /* Compute MD5 digest of space-separated pair
     * of recursive digest and interface name.
     */
    MD5Init(&ctx);
    MD5Update(&ctx, (vrpsx_uint8_t const *)digest_text, 2 * MD5_DIGEST_LENGTH);
    MD5Update(&ctx, (vrpsx_uint8_t const *)" ", 1);
    MD5Update(&ctx, (vrpsx_uint8_t const *)name, strlen(name));
    MD5Final(pair_digest, &ctx);

    /* The last 4 octets of the digest of the pair is the 32-bit big-endian
     * unsigned integer representation of the interface identifier.
     */
    return ((vrpsx_uint32_t)pair_digest[MD5_DIGEST_LENGTH - 4] << 24)
         | ((vrpsx_uint32_t)pair_digest[MD5_DIGEST_LENGTH - 3] << 16)
         | ((vrpsx_uint32_t)pair_digest[MD5_DIGEST_LENGTH - 2] << 8)
         | ((vrpsx_uint32_t)pair_digest[MD5_DIGEST_LENGTH - 1]);
}

int interface_identifier(
    char **error_message,
    char const *const *include_directories,
    char const *path,
    char const *name,
    vrpsx_uint32_t *interface_identifier,
    vrpsx_ostreambuf_t *debug)
{
    char *qpath = NULL;
    char *next;
    recursive_digest_map_t map;
    size_t path_length;
    vrpsx_uint8_t digest[MD5_DIGEST_LENGTH];
    int ern;

    if (error_message) {
        *error_message = NULL;
    }

    *interface_identifier = 0;

    path_length = strlen(path);
    qpath = malloc(1 + path_length + 2);
    if (! qpath) {
        ern = errno;
        compose_general_error(error_message, ern);
        return ern;
    }
    qpath[0] = '"';
    next = qpath + 1;
    MEMCPY_END(next, path, path_length);
    *next++ = '"';
    *next = '\0';

    recursive_digest_map_init(&map);

    ern = recursive_digest(
        error_message,
        &map,
        include_directories,
        NULL,
        0,
        ".",
        qpath,
        digest,
        debug);

    recursive_digest_map_destroy(&map);

    free(qpath);
    qpath = NULL;

    if (ern) {
        return ern;
    }

    *interface_identifier = recursive_digest_to_interface_identifier(
        digest,
        name);
    return 0;
}


void interface_identifier_to_string(
    vrpsx_uint32_t interface_identifier,
    char interface_identifier_string[9])
{
    /* When expressed as a string, we use the lowercase
     * hexadecimal representation with exactly 8 digits.
     */
    char *cp = interface_identifier_string + 8;
    *cp = '\0';
    while (cp != interface_identifier_string) {
        *--cp = xdigit[interface_identifier & 0xF];
        interface_identifier >>= 4;
    }
}
